# Generated by Django 2.2.7 on 2019-11-27 23:50

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('core', '0017_remove_task_parent'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='publication',
            name='publisher',
        ),
        migrations.DeleteModel(
            name='Publisher',
        ),
    ]
