from pulpcore.app.role_util import (  # noqa
    assign_role,
    get_groups_with_perms,
    get_groups_with_perms_attached_perms,
    get_groups_with_perms_attached_roles,
    get_objects_for_group,
    get_objects_for_user,
    get_perms_for_model,
    get_users_with_perms,
    get_users_with_perms_attached_perms,
    get_users_with_perms_attached_roles,
    remove_role,
)
