# Generated by Django 2.2.24 on 2021-06-21 19:20

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('core', '0066_download_concurrency_and_retry_changes'),
    ]

    operations = [
        migrations.AlterField(
            model_name='taskreservedresource',
            name='resource',
            field=models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, to='core.ReservedResource'),
        ),
    ]
