# Generated by Django 3.2.8 on 2021-10-21 10:58

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('core', '0081_reapplabel_group_permissions'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='task',
            options={'permissions': [('manage_roles_task', 'Can manage role assignments on task')]},
        ),
    ]
