# Generated by Django 2.2.13 on 2020-07-14 20:57

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('core', '0037_pulptemporaryfile'),
    ]

    operations = [
        migrations.AddField(
            model_name='repository',
            name='remote',
            field=models.ForeignKey(null=True, on_delete=django.db.models.deletion.SET_NULL, to='core.Remote'),
        ),
    ]
