# Generated by Django 2.2.6 on 2019-10-11 17:29

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('core', '0010_pulp_fields'),
    ]

    operations = [
        migrations.AlterField(
            model_name='contentartifact',
            name='relative_path',
            field=models.TextField(),
        ),
        migrations.AlterField(
            model_name='publishedartifact',
            name='relative_path',
            field=models.TextField(),
        ),
        migrations.AlterField(
            model_name='publishedmetadata',
            name='relative_path',
            field=models.TextField(),
        ),
    ]
