# Generated by Django 2.2.20 on 2021-04-14 14:44

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('core', '0062_add_new_distribution_mastermodel'),
    ]

    operations = [
        migrations.AddField(
            model_name='repository',
            name='retained_versions',
            field=models.PositiveIntegerField(default=None, null=True),
        ),
    ]
