# Generated by Django 3.2.9 on 2021-12-06 16:28

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('core', '0082_add_manage_roles_permissions'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='group',
            options={'permissions': [('manage_roles_group', 'Can manage role assignments on group')]},
        ),
    ]
