# Generated by Django 2.2.6 on 2019-11-04 20:00

from django.db import migrations, models
import uuid


class Migration(migrations.Migration):

    dependencies = [
        ('core', '0011_relative_path'),
    ]

    operations = [
        migrations.CreateModel(
            name='FileSystemExporter',
            fields=[
                ('pulp_id', models.UUIDField(default=uuid.uuid4, editable=False, primary_key=True, serialize=False)),
                ('pulp_created', models.DateTimeField(auto_now_add=True)),
                ('pulp_last_updated', models.DateTimeField(auto_now=True, null=True)),
                ('pulp_type', models.TextField(default=None)),
                ('name', models.CharField(db_index=True, max_length=255, unique=True)),
                ('path', models.TextField()),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.DeleteModel(
            name='Exporter',
        ),
    ]
