# Generated by Django 2.2.5 on 2019-09-18 15:52

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('core', '0007_delete_progress_proxies'),
    ]

    operations = [
        migrations.AlterUniqueTogether(
            name='publishedmetadata',
            unique_together={('publication', 'relative_path')},
        ),
        migrations.RemoveField(
            model_name='publishedmetadata',
            name='_created',
        ),
        migrations.RemoveField(
            model_name='publishedmetadata',
            name='_id',
        ),
        migrations.RemoveField(
            model_name='publishedmetadata',
            name='_last_updated',
        ),
        migrations.RemoveField(
            model_name='publishedmetadata',
            name='file',
        ),
        migrations.AddField(
            model_name='publishedmetadata',
            name='content_ptr',
            field=models.OneToOneField(auto_created=True, default=0,
                                       on_delete=django.db.models.deletion.CASCADE,
                                       parent_link=True, primary_key=True,
                                       related_name='published_metadata',
                                       serialize=False, to='core.Content'),
            preserve_default=False,
        ),
    ]
