# Generated by Django 2.2.2 on 2019-07-05 17:06

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('core', '0001_initial'),
    ]

    operations = [
        migrations.AlterField(
            model_name='artifact',
            name='size',
            field=models.BigIntegerField(),
        ),
    ]
