# Generated by Django 2.2.6 on 2019-10-01 19:31

from django.db import migrations, models
import django.db.models.deletion
import pulpcore.app.models.task


class Migration(migrations.Migration):

    dependencies = [
        ('core', '0009_remove_task_non_fatal_errors'),
    ]

    operations = [
        migrations.RenameField(
            model_name='artifact',
            old_name='_created',
            new_name='pulp_created',
        ),
        migrations.RenameField(
            model_name='artifact',
            old_name='_id',
            new_name='pulp_id',
        ),
        migrations.RenameField(
            model_name='artifact',
            old_name='_last_updated',
            new_name='pulp_last_updated',
        ),
        migrations.RenameField(
            model_name='basedistribution',
            old_name='_created',
            new_name='pulp_created',
        ),
        migrations.RenameField(
            model_name='basedistribution',
            old_name='_id',
            new_name='pulp_id',
        ),
        migrations.RenameField(
            model_name='basedistribution',
            old_name='_last_updated',
            new_name='pulp_last_updated',
        ),
        migrations.RenameField(
            model_name='basedistribution',
            old_name='_type',
            new_name='pulp_type',
        ),
        migrations.RenameField(
            model_name='content',
            old_name='_created',
            new_name='pulp_created',
        ),
        migrations.RenameField(
            model_name='content',
            old_name='_id',
            new_name='pulp_id',
        ),
        migrations.RenameField(
            model_name='content',
            old_name='_last_updated',
            new_name='pulp_last_updated',
        ),
        migrations.RenameField(
            model_name='content',
            old_name='_type',
            new_name='pulp_type',
        ),
        migrations.RenameField(
            model_name='contentappstatus',
            old_name='_created',
            new_name='pulp_created',
        ),
        migrations.RenameField(
            model_name='contentappstatus',
            old_name='_id',
            new_name='pulp_id',
        ),
        migrations.RenameField(
            model_name='contentappstatus',
            old_name='_last_updated',
            new_name='pulp_last_updated',
        ),
        migrations.RenameField(
            model_name='contentartifact',
            old_name='_created',
            new_name='pulp_created',
        ),
        migrations.RenameField(
            model_name='contentartifact',
            old_name='_id',
            new_name='pulp_id',
        ),
        migrations.RenameField(
            model_name='contentartifact',
            old_name='_last_updated',
            new_name='pulp_last_updated',
        ),
        migrations.RenameField(
            model_name='contentguard',
            old_name='_created',
            new_name='pulp_created',
        ),
        migrations.RenameField(
            model_name='contentguard',
            old_name='_id',
            new_name='pulp_id',
        ),
        migrations.RenameField(
            model_name='contentguard',
            old_name='_last_updated',
            new_name='pulp_last_updated',
        ),
        migrations.RenameField(
            model_name='contentguard',
            old_name='_type',
            new_name='pulp_type',
        ),
        migrations.RenameField(
            model_name='createdresource',
            old_name='_created',
            new_name='pulp_created',
        ),
        migrations.RenameField(
            model_name='createdresource',
            old_name='_id',
            new_name='pulp_id',
        ),
        migrations.RenameField(
            model_name='createdresource',
            old_name='_last_updated',
            new_name='pulp_last_updated',
        ),
        migrations.RenameField(
            model_name='exporter',
            old_name='_created',
            new_name='pulp_created',
        ),
        migrations.RenameField(
            model_name='exporter',
            old_name='_id',
            new_name='pulp_id',
        ),
        migrations.RenameField(
            model_name='exporter',
            old_name='_last_updated',
            new_name='pulp_last_updated',
        ),
        migrations.RenameField(
            model_name='exporter',
            old_name='_type',
            new_name='pulp_type',
        ),
        migrations.RenameField(
            model_name='progressreport',
            old_name='_created',
            new_name='pulp_created',
        ),
        migrations.RenameField(
            model_name='progressreport',
            old_name='_id',
            new_name='pulp_id',
        ),
        migrations.RenameField(
            model_name='progressreport',
            old_name='_last_updated',
            new_name='pulp_last_updated',
        ),
        migrations.RenameField(
            model_name='publication',
            old_name='_created',
            new_name='pulp_created',
        ),
        migrations.RenameField(
            model_name='publication',
            old_name='_id',
            new_name='pulp_id',
        ),
        migrations.RenameField(
            model_name='publication',
            old_name='_last_updated',
            new_name='pulp_last_updated',
        ),
        migrations.RenameField(
            model_name='publication',
            old_name='_type',
            new_name='pulp_type',
        ),
        migrations.RenameField(
            model_name='publishedartifact',
            old_name='_created',
            new_name='pulp_created',
        ),
        migrations.RenameField(
            model_name='publishedartifact',
            old_name='_id',
            new_name='pulp_id',
        ),
        migrations.RenameField(
            model_name='publishedartifact',
            old_name='_last_updated',
            new_name='pulp_last_updated',
        ),
        migrations.RenameField(
            model_name='publisher',
            old_name='_created',
            new_name='pulp_created',
        ),
        migrations.RenameField(
            model_name='publisher',
            old_name='_id',
            new_name='pulp_id',
        ),
        migrations.RenameField(
            model_name='publisher',
            old_name='_last_updated',
            new_name='pulp_last_updated',
        ),
        migrations.RenameField(
            model_name='publisher',
            old_name='_type',
            new_name='pulp_type',
        ),
        migrations.RenameField(
            model_name='remote',
            old_name='_created',
            new_name='pulp_created',
        ),
        migrations.RenameField(
            model_name='remote',
            old_name='_id',
            new_name='pulp_id',
        ),
        migrations.RenameField(
            model_name='remote',
            old_name='_last_updated',
            new_name='pulp_last_updated',
        ),
        migrations.RenameField(
            model_name='remote',
            old_name='_type',
            new_name='pulp_type',
        ),
        migrations.RenameField(
            model_name='remoteartifact',
            old_name='_created',
            new_name='pulp_created',
        ),
        migrations.RenameField(
            model_name='remoteartifact',
            old_name='_id',
            new_name='pulp_id',
        ),
        migrations.RenameField(
            model_name='remoteartifact',
            old_name='_last_updated',
            new_name='pulp_last_updated',
        ),
        migrations.RenameField(
            model_name='repository',
            old_name='_created',
            new_name='pulp_created',
        ),
        migrations.RenameField(
            model_name='repository',
            old_name='_id',
            new_name='pulp_id',
        ),
        migrations.RenameField(
            model_name='repository',
            old_name='_last_updated',
            new_name='pulp_last_updated',
        ),
        migrations.RenameField(
            model_name='repositorycontent',
            old_name='_created',
            new_name='pulp_created',
        ),
        migrations.RenameField(
            model_name='repositorycontent',
            old_name='_id',
            new_name='pulp_id',
        ),
        migrations.RenameField(
            model_name='repositorycontent',
            old_name='_last_updated',
            new_name='pulp_last_updated',
        ),
        migrations.RenameField(
            model_name='repositoryversion',
            old_name='_created',
            new_name='pulp_created',
        ),
        migrations.RenameField(
            model_name='repositoryversion',
            old_name='_id',
            new_name='pulp_id',
        ),
        migrations.RenameField(
            model_name='repositoryversion',
            old_name='_last_updated',
            new_name='pulp_last_updated',
        ),
        migrations.RenameField(
            model_name='reservedresource',
            old_name='_created',
            new_name='pulp_created',
        ),
        migrations.RenameField(
            model_name='reservedresource',
            old_name='_id',
            new_name='pulp_id',
        ),
        migrations.RenameField(
            model_name='reservedresource',
            old_name='_last_updated',
            new_name='pulp_last_updated',
        ),
        migrations.RenameField(
            model_name='reservedresourcerecord',
            old_name='_created',
            new_name='pulp_created',
        ),
        migrations.RenameField(
            model_name='reservedresourcerecord',
            old_name='_id',
            new_name='pulp_id',
        ),
        migrations.RenameField(
            model_name='reservedresourcerecord',
            old_name='_last_updated',
            new_name='pulp_last_updated',
        ),
        migrations.RenameField(
            model_name='task',
            old_name='_created',
            new_name='pulp_created',
        ),
        migrations.RenameField(
            model_name='task',
            old_name='_id',
            new_name='pulp_id',
        ),
        migrations.RenameField(
            model_name='task',
            old_name='_last_updated',
            new_name='pulp_last_updated',
        ),
        migrations.RenameField(
            model_name='taskreservedresource',
            old_name='_created',
            new_name='pulp_created',
        ),
        migrations.RenameField(
            model_name='taskreservedresource',
            old_name='_id',
            new_name='pulp_id',
        ),
        migrations.RenameField(
            model_name='taskreservedresource',
            old_name='_last_updated',
            new_name='pulp_last_updated',
        ),
        migrations.RenameField(
            model_name='taskreservedresourcerecord',
            old_name='_created',
            new_name='pulp_created',
        ),
        migrations.RenameField(
            model_name='taskreservedresourcerecord',
            old_name='_id',
            new_name='pulp_id',
        ),
        migrations.RenameField(
            model_name='taskreservedresourcerecord',
            old_name='_last_updated',
            new_name='pulp_last_updated',
        ),
        migrations.RenameField(
            model_name='upload',
            old_name='_created',
            new_name='pulp_created',
        ),
        migrations.RenameField(
            model_name='upload',
            old_name='_id',
            new_name='pulp_id',
        ),
        migrations.RenameField(
            model_name='upload',
            old_name='_last_updated',
            new_name='pulp_last_updated',
        ),
        migrations.RenameField(
            model_name='uploadchunk',
            old_name='_created',
            new_name='pulp_created',
        ),
        migrations.RenameField(
            model_name='uploadchunk',
            old_name='_id',
            new_name='pulp_id',
        ),
        migrations.RenameField(
            model_name='uploadchunk',
            old_name='_last_updated',
            new_name='pulp_last_updated',
        ),
        migrations.RenameField(
            model_name='worker',
            old_name='_created',
            new_name='pulp_created',
        ),
        migrations.RenameField(
            model_name='worker',
            old_name='_id',
            new_name='pulp_id',
        ),
        migrations.RenameField(
            model_name='worker',
            old_name='_last_updated',
            new_name='pulp_last_updated',
        ),
        migrations.AlterField(
            model_name='basedistribution',
            name='content_guard',
            field=models.ForeignKey(null=True, on_delete=django.db.models.deletion.SET_NULL, to='core.ContentGuard', to_field='pulp_id'),
        ),
        migrations.AlterField(
            model_name='basedistribution',
            name='remote',
            field=models.ForeignKey(null=True, on_delete=django.db.models.deletion.SET_NULL, to='core.Remote', to_field='pulp_id'),
        ),
        migrations.AlterField(
            model_name='contentartifact',
            name='artifact',
            field=models.ForeignKey(null=True, on_delete=django.db.models.deletion.PROTECT, to='core.Artifact', to_field='pulp_id'),
        ),
        migrations.AlterField(
            model_name='contentartifact',
            name='content',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='core.Content', to_field='pulp_id'),
        ),
        migrations.AlterField(
            model_name='createdresource',
            name='task',
            field=models.ForeignKey(default=pulpcore.app.models.task.Task.current, on_delete=django.db.models.deletion.CASCADE, related_name='created_resources', to='core.Task', to_field='pulp_id'),
        ),
        migrations.AlterField(
            model_name='progressreport',
            name='task',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='progress_reports', to='core.Task', to_field='pulp_id'),
        ),
        migrations.AlterField(
            model_name='publication',
            name='publisher',
            field=models.ForeignKey(null=True, on_delete=django.db.models.deletion.CASCADE, to='core.Publisher', to_field='pulp_id'),
        ),
        migrations.AlterField(
            model_name='publication',
            name='repository_version',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='core.RepositoryVersion', to_field='pulp_id'),
        ),
        migrations.AlterField(
            model_name='publishedartifact',
            name='content_artifact',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='published_artifact', to='core.ContentArtifact', to_field='pulp_id'),
        ),
        migrations.AlterField(
            model_name='publishedartifact',
            name='publication',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='published_artifact', to='core.Publication', to_field='pulp_id'),
        ),
        migrations.AlterField(
            model_name='publishedmetadata',
            name='content_ptr',
            field=models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, related_name='published_metadata', serialize=False, to='core.Content', to_field='pulp_id'),
        ),
        migrations.AlterField(
            model_name='publishedmetadata',
            name='publication',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='published_metadata', to='core.Publication', to_field='pulp_id'),
        ),
        migrations.AlterField(
            model_name='remoteartifact',
            name='content_artifact',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='core.ContentArtifact', to_field='pulp_id'),
        ),
        migrations.AlterField(
            model_name='remoteartifact',
            name='remote',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='core.Remote', to_field='pulp_id'),
        ),
        migrations.AlterField(
            model_name='repositorycontent',
            name='content',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='version_memberships', to='core.Content', to_field='pulp_id'),
        ),
        migrations.AlterField(
            model_name='repositorycontent',
            name='repository',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='core.Repository', to_field='pulp_id'),
        ),
        migrations.AlterField(
            model_name='repositorycontent',
            name='version_added',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='added_memberships', to='core.RepositoryVersion', to_field='pulp_id'),
        ),
        migrations.AlterField(
            model_name='repositorycontent',
            name='version_removed',
            field=models.ForeignKey(null=True, on_delete=django.db.models.deletion.CASCADE, related_name='removed_memberships', to='core.RepositoryVersion', to_field='pulp_id'),
        ),
        migrations.AlterField(
            model_name='repositoryversion',
            name='base_version',
            field=models.ForeignKey(null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='versions', to='core.RepositoryVersion', to_field='pulp_id'),
        ),
        migrations.AlterField(
            model_name='repositoryversion',
            name='repository',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='versions', to='core.Repository', to_field='pulp_id'),
        ),
        migrations.AlterField(
            model_name='repositoryversioncontentdetails',
            name='repository_version',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='counts', to='core.RepositoryVersion', to_field='pulp_id'),
        ),
        migrations.AlterField(
            model_name='reservedresource',
            name='worker',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='reservations', to='core.Worker', to_field='pulp_id'),
        ),
        migrations.AlterField(
            model_name='task',
            name='parent',
            field=models.ForeignKey(null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='spawned_tasks', to='core.Task', to_field='pulp_id'),
        ),
        migrations.AlterField(
            model_name='task',
            name='worker',
            field=models.ForeignKey(null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='tasks', to='core.Worker', to_field='pulp_id'),
        ),
        migrations.AlterField(
            model_name='taskreservedresource',
            name='resource',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='core.ReservedResource', to_field='pulp_id'),
        ),
        migrations.AlterField(
            model_name='taskreservedresource',
            name='task',
            field=models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, to='core.Task', to_field='pulp_id'),
        ),
        migrations.AlterField(
            model_name='taskreservedresourcerecord',
            name='resource',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='core.ReservedResourceRecord', to_field='pulp_id'),
        ),
        migrations.AlterField(
            model_name='taskreservedresourcerecord',
            name='task',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='core.Task', to_field='pulp_id'),
        ),
        migrations.AlterField(
            model_name='uploadchunk',
            name='upload',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='chunks', to='core.Upload', to_field='pulp_id'),
        ),
    ]
