# Generated by Django 2.2.11 on 2020-04-07 15:47

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('core', '0026_task_group'),
    ]

    operations = [
        migrations.CreateModel(
            name='PulpExport',
            fields=[
                ('export_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, related_name='core_pulp_export', serialize=False, to='core.Export')),
                ('sha256', models.CharField(max_length=64, null=True)),
                ('filename', models.CharField(max_length=4096, null=True)),
            ],
            options={
                'default_related_name': '%(app_label)s_pulp_export',
            },
            bases=('core.export',),
        ),
        migrations.AlterField(
            model_name='pulpexporter',
            name='last_export',
            field=models.ForeignKey(null=True, on_delete=django.db.models.deletion.PROTECT, related_name='last_export', to='core.PulpExport'),
        ),
    ]
