# Generated by Django 2.2.17 on 2020-11-12 19:38

import django.core.validators
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('core', '0050_namespace_access_policies'),
    ]

    operations = [
        migrations.AddField(
            model_name='remote',
            name='connect_timeout',
            field=models.FloatField(null=True, validators=[django.core.validators.MinValueValidator(0.0, 'Timeout must be >= 0')]),
        ),
        migrations.AddField(
            model_name='remote',
            name='sock_connect_timeout',
            field=models.FloatField(null=True, validators=[django.core.validators.MinValueValidator(0.0, 'Timeout must be >= 0')]),
        ),
        migrations.AddField(
            model_name='remote',
            name='sock_read_timeout',
            field=models.FloatField(null=True, validators=[django.core.validators.MinValueValidator(0.0, 'Timeout must be >= 0')]),
        ),
        migrations.AddField(
            model_name='remote',
            name='total_timeout',
            field=models.FloatField(null=True, validators=[django.core.validators.MinValueValidator(0.0, 'Timeout must be >= 0')]),
        ),
    ]
