# Generated by Django 2.2.12 on 2020-05-04 04:41

from django.db import migrations, models


def set_existing_task_groups_finished(apps, schema_editor):
    TaskGroup = apps.get_model('core', 'TaskGroup')
    TaskGroup.objects.all().update(all_tasks_dispatched=True)


class Migration(migrations.Migration):

    dependencies = [
        ('core', '0029_export_delete'),
    ]

    operations = [
        migrations.AddField(
            model_name='taskgroup',
            name='all_tasks_dispatched',
            field=models.BooleanField(default=False),
        ),
        migrations.RunPython(set_existing_task_groups_finished, reverse_code=migrations.RunPython.noop, elidable=True),
    ]
