# Generated by Django 2.2.24 on 2021-06-13 03:14

import django.core.validators
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('core', '0065_merge_20210615_1211'),
    ]

    operations = [
        migrations.AlterField(
            model_name='remote',
            name='download_concurrency',
            field=models.PositiveIntegerField(null=True, validators=[django.core.validators.MinValueValidator(1, 'Download concurrency must be at least 1')]),
        ),
        migrations.AddField(
            model_name='remote',
            name='max_retries',
            field=models.PositiveIntegerField(null=True),
        ),
    ]
