# Generated by Django 2.2.6 on 2019-10-07 16:45

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('core', '0008_published_metadata_as_content'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='task',
            name='non_fatal_errors',
        ),
    ]
