# Generated by Django 2.2.5 on 2019-09-17 16:59

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('core', '0006_repository_plugin_managed'),
    ]

    operations = [
        migrations.DeleteModel(
            name='ProgressBar',
        ),
        migrations.DeleteModel(
            name='ProgressSpinner',
        ),
    ]
