# Generated by Django 2.2.7 on 2019-11-11 16:48

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('core', '0013_repository_pulp_type'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='repository',
            name='plugin_managed',
        ),
    ]
