# Generated by Django 3.2.7 on 2021-09-22 12:55

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('core', '0078_grouprole_role_userrole'),
    ]

    operations = [
        migrations.RenameField(
            model_name='accesspolicy',
            old_name='permissions_assignment',
            new_name='creation_hooks',
        ),
    ]
