# Generated by Django 2.2.12 on 2020-06-23 13:57

from django.db import migrations, models
import django.db.models.deletion
import uuid


class Migration(migrations.Migration):

    dependencies = [
        ('core', '0033_increase_remote_artifact_size_field'),
    ]

    operations = [
        migrations.CreateModel(
            name='GroupProgressReport',
            fields=[
                ('pulp_id', models.UUIDField(default=uuid.uuid4, editable=False, primary_key=True, serialize=False)),
                ('pulp_created', models.DateTimeField(auto_now_add=True)),
                ('pulp_last_updated', models.DateTimeField(auto_now=True, null=True)),
                ('message', models.TextField()),
                ('code', models.CharField(max_length=36)),
                ('total', models.IntegerField(default=0)),
                ('done', models.IntegerField(default=0)),
                ('suffix', models.TextField(null=True)),
                ('task_group', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='group_progress_reports', to='core.TaskGroup')),
            ],
            options={
                'unique_together': {('code', 'task_group')},
            },
        ),
    ]
