# Generated by Django 2.2.6 on 2019-10-28 18:42

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('core', '0012_auto_20191104_2000'),
    ]

    operations = [
        migrations.AddField(
            model_name='repository',
            name='pulp_type',
            field=models.TextField(default=None),
        ),
    ]
