# Generated by Django 2.2.14 on 2020-07-15 14:44

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('core', '0035_content_upstream_id'),
    ]

    operations = [
        migrations.AlterField(
            model_name='pulpexporter',
            name='last_export',
            field=models.ForeignKey(null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='last_export', to='core.PulpExport'),
        ),
    ]
