# Generated by Django 2.2.17 on 2021-01-29 19:17

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('core', '0056_remote_rate_limit'),
    ]

    operations = [
        migrations.AlterField(
            model_name='label',
            name='key',
            field=models.CharField(db_index=True, max_length=200),
        ),
        migrations.AlterField(
            model_name='label',
            name='value',
            field=models.TextField(db_index=True, null=True),
        ),
    ]
