# Generated by Django 2.2.15 on 2020-08-25 19:32

from django.db import migrations
import pulpcore.app.models.content
import pulpcore.app.models.fields


class Migration(migrations.Migration):

    dependencies = [
        ('core', '0043_toc_attribute'),
    ]

    operations = [
        migrations.AlterField(
            model_name='pulptemporaryfile',
            name='file',
            field=pulpcore.app.models.fields.ArtifactFileField(max_length=255, upload_to=pulpcore.app.models.content.PulpTemporaryFile.storage_path),
        ),
    ]
