# Generated by Django 3.2.8 on 2021-12-09 22:59

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('core', '0083_alter_group_options'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='rbaccontentguard',
            options={'default_related_name': '%(app_label)s_%(model_name)s', 'permissions': (('download_rbaccontentguard', 'Can Download Content'), ('manage_roles_rbaccontentguard', 'Can manage role assignments on RBAC content guard'))},
        ),
    ]
